// ignore_for_file: avoid_print, non_constant_identifier_names

import 'dart:convert';
import 'dart:io';
import 'dart:math';
import 'dart:typed_data';

import 'package:args/args.dart';
import 'package:encrypt/encrypt.dart';
import 'package:crypto/crypto.dart';

// ignore: avoid_relative_lib_imports
import '../lib/src/encrypt.dart';
// ignore: avoid_relative_lib_imports
import '../lib/src/config.dart';

// ignore: constant_identifier_names
const String AssetManifestKey = "AssetManifest.json";
// ignore: constant_identifier_names
const String FontManifestKey = "FontManifest.json";

/// flutter_assets 排除名单
List<String> VBDL_EXCLUDED_ITEMS = [
  "LICENSE",
  "isolate_snapshot_data",
  "kernel_blob.bin",
  "vm_snapshot_data",
];

List<String> VBDL_EXCLUDED_FONTS = [

];

void quit(int code, [String? message]) {
  var prefix = "";

  if (code != 0) prefix = "error:";
  if (message != null) {
    print("$prefix $message");
  }
  exit(code);
}

void entrypoint(List<String> arguments) async {
  final ArgParser argParser = ArgParser()
    ..addOption("dir", abbr: "d", help: "资源文件夹路径")
    ..addOption("id", abbr: "i", help: "应用ID")
    ..addOption("exc", abbr: "e", help: "白名单文件")
    ..addFlag("help", abbr: "h", negatable: false, help: "帮助");
  final ArgResults argResults = argParser.parse(arguments);

  if (argResults["help"]) {
    print("""
${argParser.usage}

""");
    return quit(0);
  }

  String? defaultProjectDirectory() {
    return (Platform.environment["VBDL_PROJECT_DIR"] ?? Platform.environment["PROJECT_DIR"]) ?? "";
  }

  String getKeyForAppId(String appid) {
    final rawKey = "$VBDL_HEADER_ENCKEY.$appid.$VBDL_DATA_ENCKEY";
    return md5.convert(utf8.encode(rawKey)).toString();
  }

  String? path = argResults["dir"];
  if (path == null) {
    final projDir = defaultProjectDirectory();
    if (projDir != null) {
      // path = "$projDir/Flutter/App.framework/flutter_assets";
      path = "${Platform.environment["FLUTTER_ASSET_DIR_NAME"] ?? ""}";
      // /Users/jiangao/work/Dream/newfluttercgapp/ios  /Flutter/App.framework/flutter_assets
      // "$projDir/../build/flutter_assets";
      print("----1----, VBDL_PROJECT_DIR: ${Platform.environment["VBDL_PROJECT_DIR"] ?? ""}");
      print("-----2---, PROJECT_DIR: ${Platform.environment["PROJECT_DIR"] ?? ""}");
      print("----3----,  $path");
      print("-----4---, FLUTTER_ASSET_DIR_NAME: ${Platform.environment["FLUTTER_ASSET_DIR_NAME"] ?? ""}");
    }
  }
  if (path == null) {
    return quit(-1, "dir is required.");
  }
  final dir = Directory(path);
  if (dir.existsSync() != true) {
    return quit(-1, "$path does not exist.");
  }
  if (path.endsWith("/flutter_assets") != true) {
    return quit(-1, "dir must be \"flutter_assets\".");
  }
  
  String? appid = argResults["id"] ?? Platform.environment["PRODUCT_BUNDLE_IDENTIFIER"];
  
  if (appid == null) {
    return quit(-1, "id is required.");
  }

  final String key = getKeyForAppId(appid);

  String? exc = argResults["exc"];
  if (exc == null) {
    final projDir = defaultProjectDirectory();
    if (projDir != null) {
      exc = "$projDir/FLTVBDLExcluded.json";
    }
  }
  if (exc != null && File(exc).existsSync()) {
    final items = (json.decode(File(exc).readAsStringSync()) as Map).cast<String, List>();
    VBDL_EXCLUDED_ITEMS.addAll((items["assets"] ?? []).cast<String>());
    VBDL_EXCLUDED_FONTS = (items["fonts"] ?? []).cast<String>();
  }

  final excluded = <String>[];
  for (final ex in VBDL_EXCLUDED_ITEMS) {
    excluded.add([dir.path, ex].join("/"));
  }
  VBDL_EXCLUDED_ITEMS = excluded;

  print("[VBDL] Pack => key:$key dir:${dir.path}");
  await VBDLTool(dir.path, key, [dir.path, VBDL_ASSETS_KEY].join("/")).pack();
  print("[VBDL] SUCCESS");

  return quit(0);
}

void main(List<String> arguments) async {
  // arguments = ["-d", "/Users/gumingjun/Downloads/A/bk/flutter_assets", "-k", "qwer1234", "-e", "/Users/gumingjun/Downloads/A/FLTVBDLExcluded.json"];
  // print("$arguments");
  entrypoint(arguments);
}

class VBDLTool {
  final Directory rootDir;
  late Map<String, dynamic> _orignalAssetManifest;
  late List<Map<String, dynamic>> _orignalFontManifest;
  File _outputFile;

  late Key _headerDecKey;
  late Key _dataDecKey;
  late Key _indexDecKey;

  Map<String, List<int>> _indexMap = {};
  List<Uint8List> _dataList = [];

  int _dataIter = 0;

  Key _makeKey(String key) {
    final digest = md5.convert(utf8.encode(key));
    return Key(Uint8List.fromList(digest.bytes));
  }

  // late String rootPath;
  // late String key;
  // late String outputPath;

  // VBDLTool(this.rootPath, this.key, this.outputPath) 
  VBDLTool(String rootPath, String key, String outputPath)
      : rootDir = Directory(rootPath),
        _outputFile = File(outputPath) {
    _headerDecKey = _makeKey(VBDL_HEADER_ENCKEY);
    _dataDecKey = _makeKey(VBDL_DATA_ENCKEY);
    _indexDecKey = _makeKey(key);
  }

  int _random32() => Random().nextInt((1 << 32));

  int _padding16(int len) => len % 16 == 0 ? 0 : 16 - len % 16;
  int _align16(int len) => len + _padding16(len);

  Future<void> _visitDirectory(Directory dir) async {
    // 命中排除的文件夹直接返回
    if (VBDL_EXCLUDED_ITEMS.contains(dir.path)) return;
    final children = dir.listSync(followLinks: false);
    for (final child in children) {
      // 隐藏文件或文件夹直接删除
      if (child.path.split("/").last.startsWith(".")) {
        child.deleteSync(recursive: true);
        continue;
      }

      if (child is File) {
        await _visitFile(child);
      } else if (child is Directory) {
        await _visitDirectory(child);
      }
    }

    // 如果还存在子项则保留当前文件夹
    final afChildren = dir.listSync(followLinks: false);
    if (afChildren.isEmpty) {
      dir.deleteSync(recursive: true);
    }
  }

  Future<void> _visitFileDirect(File file) async {
    final key = file.path.substring(rootDir.path.length + 1);

    print(
        "packing:$key [${(file.statSync().size / 1024).toStringAsFixed(2)} KB]");
    final beginTime = DateTime.now().millisecondsSinceEpoch;

    final bytes = Uint8List.fromList(file.readAsBytesSync());
    final random = _random32();
    final encBytes = encryptData(_dataDecKey, bytes, random);
    _indexMap[key] = [_dataIter, encBytes.length, random];
    _dataList.add(encBytes);
    _dataIter += encBytes.length;
    int padding = _padding16(_dataIter);
    if (padding != 0) _dataList.add(Uint8List(padding));
    _dataIter += padding;

    final endTime = DateTime.now().millisecondsSinceEpoch;
    print("packing:$key [${(endTime - beginTime)} ms]");

    // 删除原文件
    file.deleteSync(recursive: false);
  }

  Future<void> _visitFile(File file) async {
    // 命中排除的文件直接返回
    if (VBDL_EXCLUDED_ITEMS.contains(file.path)) return;
    await _visitFileDirect(file);
  }

  void _updateManifest() {
    var afile = File([_outputFile.parent.path, AssetManifestKey].join("/"));
    final Map<String, dynamic> replacedAssetMainfest = {VBDL_ASSETS_KEY: [VBDL_ASSETS_KEY]};
    for (var key in _orignalAssetManifest.keys) {
      // 不在加密包里
      if (_indexMap[key] == null) {
        // 文件存在
        final keyf = File([afile.parent.path, key].join("/"));
        if (keyf.existsSync()) {
          /// 未加密的资源
          replacedAssetMainfest[key] = _orignalAssetManifest[key];
        } else if (!keyf.path.split("/").last.startsWith(".")) {
          print("WARNING:$key NOT MAPPING");
        }
      }
    }
    afile.writeAsStringSync(json.encode(replacedAssetMainfest));

    var bfile = File([_outputFile.parent.path, FontManifestKey].join("/"));
    final List<Map<String, dynamic>> replacedFontManifest = [];
    for (var fonts in _orignalFontManifest) {
      if (VBDL_EXCLUDED_FONTS.contains(fonts["family"] ?? "")) {
        replacedFontManifest.add(fonts);
      }
    }
    bfile.writeAsStringSync(json.encode(replacedFontManifest));
  }

  Future<void> pack() async {
    assert(rootDir.existsSync() == true);

    final beginTime = DateTime.now().millisecondsSinceEpoch;

    _dataIter = 0;

    {
      final path = [_outputFile.parent.path, AssetManifestKey].join("/");
      var file = File(path);
      _orignalAssetManifest = json.decode(file.readAsStringSync());
    }

    {
      final path = [_outputFile.parent.path, FontManifestKey].join("/");
      var file = File(path);
      _orignalFontManifest = (json.decode(file.readAsStringSync()) as List).cast<Map<String, dynamic>>();
    }

    await _visitDirectory(rootDir);

    _updateManifest();

    final encIndexData = encryptIndex(_indexDecKey, _indexMap);

    const indexStart = VBDLHeader.ENCRYPTED_LEN;
    final indexLength = encIndexData.length;

    final dataPaddingData = Uint8List(_padding16(indexStart + indexLength));
    final dataStart = _align16(indexStart + indexLength);
    final dataLength = _dataIter;

    final header =
        VBDLHeader.make(indexStart, indexLength, dataStart, dataLength);
    final encHeaderData = encryptHeader(_headerDecKey, header);
    assert(encHeaderData.length == VBDLHeader.ENCRYPTED_LEN);

    if (_outputFile.existsSync()) {
      _outputFile = await _outputFile.writeAsBytes([], flush: true);
    } else {
      _outputFile = await _outputFile.create(recursive: true);
    }

    int totalSize = 0;

    _outputFile.writeAsBytesSync(encHeaderData, mode: FileMode.append);
    totalSize += encHeaderData.length;

    _outputFile.writeAsBytesSync(encIndexData, mode: FileMode.append);
    totalSize += encIndexData.length;

    _outputFile.writeAsBytesSync(dataPaddingData, mode: FileMode.append);
    totalSize += dataPaddingData.length;

    for (var d in _dataList) {
      _outputFile.writeAsBytesSync(d, mode: FileMode.append);
      totalSize += d.length;
    }

    final filePaddingData = Uint8List(_padding16(totalSize));
    _outputFile.writeAsBytesSync(filePaddingData, mode: FileMode.append);
    totalSize += _padding16(totalSize);

    assert(_outputFile.statSync().size == totalSize);

    final endTime = DateTime.now().millisecondsSinceEpoch;
    print(
        "[OK, ${((endTime - beginTime) / 1000).toStringAsFixed(2)} s, ${(totalSize / 1024 / 1024).toStringAsFixed(2)} MB]");
  }
}
